@@ -23,7 +23,11 @@ import java.io.InputStream;
 import java.io.InputStreamReader;
 import java.io.OutputStream;
 import java.io.Reader;
+import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Collection;
+import java.util.EnumSet;
+import java.util.List;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.atomic.AtomicInteger;
 
@@ -34,19 +38,29 @@ import javax.net.ssl.SSLContext;
 import javax.net.ssl.SSLSocket;
 import javax.net.ssl.SSLSocketFactory;
 import javax.servlet.ServletException;
+import javax.servlet.SessionTrackingMode;
 import javax.servlet.http.HttpServlet;
 import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
 
 import org.junit.Assert;
-import org.junit.Assume;
 import org.junit.Test;
+import org.junit.runner.RunWith;
+import org.junit.runners.Parameterized;
+import org.junit.runners.Parameterized.Parameter;
 
 import org.apache.catalina.Context;
+import org.apache.catalina.Lifecycle;
+import org.apache.catalina.LifecycleEvent;
+import org.apache.catalina.LifecycleListener;
 import org.apache.catalina.Wrapper;
+import org.apache.catalina.connector.Connector;
+import org.apache.catalina.connector.Request;
+import org.apache.catalina.connector.Response;
 import org.apache.catalina.startup.TesterServlet;
 import org.apache.catalina.startup.Tomcat;
 import org.apache.catalina.startup.TomcatBaseTest;
+import org.apache.catalina.valves.ValveBase;
 import org.apache.tomcat.util.buf.ByteChunk;
 import org.apache.tomcat.websocket.server.WsContextListener;
 
@@ -55,8 +69,32 @@ import org.apache.tomcat.websocket.server.WsContextListener;
  * generated using a test CA the files for which are in the Tomcat PMC private
  * repository since not all of them are AL2 licensed.
  */
+@RunWith(Parameterized.class)
 public class TestSsl extends TomcatBaseTest {
 
+    @Parameterized.Parameters(name = "{0}")
+    public static Collection<Object[]> parameters() {
+        List<Object[]> parameterSets = new ArrayList<>();
+        parameterSets.add(new Object[] {
+                "JSSE", Boolean.FALSE, "org.apache.tomcat.util.net.jsse.JSSEImplementation"});
+        parameterSets.add(new Object[] {
+                "OpenSSL", Boolean.TRUE, "org.apache.tomcat.util.net.openssl.OpenSSLImplementation"});
+        parameterSets.add(new Object[] {
+                "OpenSSL-FFM", Boolean.TRUE, "org.apache.tomcat.util.net.openssl.panama.OpenSSLImplementation"});
+
+        return parameterSets;
+    }
+
+    @Parameter(0)
+    public String connectorName;
+
+    @Parameter(1)
+    public boolean useOpenSSL;
+
+    @Parameter(2)
+    public String sslImplementationName;
+
+
     @Test
     public void testSimpleSsl() throws Exception {
         TesterSupport.configureClientSsl();
@@ -68,6 +106,7 @@ public class TestSsl extends TomcatBaseTest {
         ctxt.addApplicationListener(WsContextListener.class.getName());
 
         TesterSupport.initSsl(tomcat);
+        TesterSupport.configureSSLImplementation(tomcat, sslImplementationName, useOpenSSL);
 
         tomcat.start();
         ByteChunk res = getUrl("https://localhost:" + getPort() +
@@ -85,14 +124,47 @@ public class TestSsl extends TomcatBaseTest {
 
     }
 
+    private String sessionId = null;
+
+    @Test
+    public void testSSLSessionTracking() throws Exception {
+        TesterSupport.configureClientSsl();
+
+        Tomcat tomcat = getTomcatInstance();
+        tomcat.getEngine().getPipeline().addValve(new VerificationValve());
+
+        File appDir = new File(getBuildDirectory(), "webapps/examples");
+        Context ctxt  = tomcat.addWebapp(null, "/examples", appDir.getAbsolutePath());
+        ctxt.addLifecycleListener(new SSLSessionTrackingLifecycleListener());
+        ctxt.addApplicationListener(WsContextListener.class.getName());
+
+        ctxt = getProgrammaticRootContext();
+        Tomcat.addServlet(ctxt, "post", new SimplePostServlet());
+        ctxt.addServletMappingDecoded("/post", "post");
+
+        TesterSupport.initSsl(tomcat);
+        TesterSupport.configureSSLImplementation(tomcat, sslImplementationName, useOpenSSL);
+
+        tomcat.start();
+        getUrl("https://localhost:" + getPort() + "/examples/servlets/servlet/HelloWorldExample");
+        // SSL is the only source for the requested session ID, and SessionTrackingMode.SSL is set on examples
+        Assert.assertNotNull(sessionId);
+        Assert.assertTrue(sessionId.length() > 1);
+        // No SessionTrackingMode.SSL on the root webapp
+        getUrl("https://localhost:" + getPort() + "/post");
+        Assert.assertNull(sessionId);
+    }
+
     @Test
     public void testPost() throws Exception {
         SocketFactory socketFactory = TesterSupport.configureClientSsl();
 
         Tomcat tomcat = getTomcatInstance();
         TesterSupport.initSsl(tomcat);
+        Connector connector = tomcat.getConnector();
         // Increase timeout as default (3s) can be too low for some CI systems
-        Assert.assertTrue(tomcat.getConnector().setProperty("connectionTimeout", "20000"));
+        Assert.assertTrue(connector.setProperty("connectionTimeout", "20000"));
+        TesterSupport.configureSSLImplementation(tomcat, sslImplementationName, useOpenSSL);
 
         Context ctxt = getProgrammaticRootContext();
         Tomcat.addServlet(ctxt, "post", new SimplePostServlet());
@@ -175,6 +247,8 @@ public class TestSsl extends TomcatBaseTest {
         TesterSupport.initSsl(tomcat, TesterSupport.LOCALHOST_KEYPASS_JKS,
                 TesterSupport.JKS_PASS, null, TesterSupport.JKS_KEY_PASS, null);
 
+        TesterSupport.configureSSLImplementation(tomcat, sslImplementationName, useOpenSSL);
+
         tomcat.start();
         ByteChunk res = getUrl("https://localhost:" + getPort() +
             "/examples/servlets/servlet/HelloWorldExample");
@@ -196,6 +270,8 @@ public class TestSsl extends TomcatBaseTest {
         TesterSupport.initSsl(tomcat, TesterSupport.LOCALHOST_KEYPASS_JKS,
                       null, TesterSupport.JKS_PASS_FILE, null, TesterSupport.JKS_KEY_PASS_FILE);
 
+        TesterSupport.configureSSLImplementation(tomcat, sslImplementationName, useOpenSSL);
+
         tomcat.start();
         ByteChunk res = getUrl("https://localhost:" + getPort() +
             "/examples/servlets/servlet/HelloWorldExample");
@@ -205,20 +281,19 @@ public class TestSsl extends TomcatBaseTest {
     }
 
     @Test
-    public void testRenegotiateWorks() throws Exception {
-        Tomcat tomcat = getTomcatInstance();
+    public void testClientInitiatedRenegotiation() throws Exception {
 
-        Assume.assumeTrue("SSL renegotiation has to be supported for this test",
-                TesterSupport.isClientRenegotiationSupported(getTomcatInstance()));
+        Tomcat tomcat = getTomcatInstance();
+        TesterSupport.initSsl(tomcat);
+        TesterSupport.configureSSLImplementation(tomcat, sslImplementationName, useOpenSSL);
+
+        boolean renegotiationSupported = TesterSupport.isClientRenegotiationSupported(getTomcatInstance());
 
         Context root = tomcat.addContext("", TEMP_DIR);
-        Wrapper w =
-            Tomcat.addServlet(root, "tester", new TesterServlet());
+        Wrapper w = Tomcat.addServlet(root, "tester", new TesterServlet());
         w.setAsyncSupported(true);
         root.addServletMappingDecoded("/", "tester");
 
-        TesterSupport.initSsl(tomcat);
-
         tomcat.start();
 
         SSLContext sslCtx;
@@ -232,8 +307,7 @@ public class TestSsl extends TomcatBaseTest {
         }
         sslCtx.init(null, TesterSupport.getTrustManagers(), null);
         SSLSocketFactory socketFactory = sslCtx.getSocketFactory();
-        SSLSocket socket = (SSLSocket) socketFactory.createSocket("localhost",
-                getPort());
+        SSLSocket socket = (SSLSocket) socketFactory.createSocket("localhost", getPort());
 
         OutputStream os = socket.getOutputStream();
         InputStream is = socket.getInputStream();
@@ -248,7 +322,17 @@ public class TestSsl extends TomcatBaseTest {
 
         socket.startHandshake();
 
-        doRequest(os, r);
+        try {
+            doRequest(os, r);
+            if (!renegotiationSupported) {
+                Assert.fail("Renegotiation started when it should have failed");
+            }
+        } catch (IOException e) {
+            if (renegotiationSupported) {
+                Assert.fail("Renegotiation failed when it should be supported");
+            }
+            return;
+        }
         // Handshake complete appears to be called asynchronously
         int wait = 0;
         while (wait < 5000 && !listener.isComplete()) {
@@ -336,4 +420,24 @@ public class TestSsl extends TomcatBaseTest {
         }
 
     }
+
+    public static class SSLSessionTrackingLifecycleListener implements LifecycleListener {
+        @Override
+        public void lifecycleEvent(LifecycleEvent event) {
+            if (Lifecycle.BEFORE_START_EVENT.equals(event.getType())) {
+                if (!(event.getLifecycle() instanceof Context)) {
+                    throw new IllegalArgumentException("Not Context");
+                }
+                Context context = (Context) event.getLifecycle();
+                context.getServletContext().setSessionTrackingModes(EnumSet.of(SessionTrackingMode.SSL));
+            }
+        }
+    }
+
+    public class VerificationValve extends ValveBase {
+        @Override
+        public void invoke(Request request, Response response) throws IOException, ServletException {
+            sessionId = request.getRequestedSessionId();
+        }
+    }
 }
